from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Mapping, Type

from pydantic import BaseModel

from ...types import FormatLabel

if TYPE_CHECKING:
    from .formatter import FormatterInterface


class FormatPackage(BaseModel):
    label: FormatLabel
    data: Mapping[str, Any]


def register_format(label: FormatLabel):
    def decorator(cls: Type[FormatterInterface]):
        FormatDispatcher.register(label, cls)
        setattr(cls, "__registered_format__", label)
        return cls

    return decorator


class FormatDispatcher:
    _registered_formats: Dict[FormatLabel, Type[FormatterInterface]] = {}

    @classmethod
    def register(cls, label: FormatLabel, formatter: Type[FormatterInterface]) -> None:
        cls._registered_formats[label] = formatter

    @classmethod
    def get(cls, label: FormatLabel) -> Type[FormatterInterface]:
        return cls._registered_formats[label]

    @classmethod
    def format(cls, package: FormatPackage) -> str:
        label = package.label
        data = package.data
        return cls.get(label).to_string(data)



from pathlib import Path


class PathConfig:
    DEFAULT_ROOT_PATH: Path = Path("/autograder")
    DEFAULT_SOURCE_PATH: Path = Path("/autograder/source")
    DEFAULT_SUBMISSION_PATH: Path = Path("/autograder/submission")
    DEFAULT_RESULT_PATH: Path = Path("/autograder/results/results.json")



import subprocess
from collections import deque
from pathlib import Path

from ..types import ProjectType


def detect_project_type(project_root: Path) -> ProjectType:
    for file in bfs_walk(project_root):
        if is_cmake_file(file):
            return "cmake"
        if is_makefile_file(file):
            return "makefile"
    return "cxx-source"


def bfs_walk(root: Path):  # pathlib defaults to dfs; must implement bfs ourselves.
    queue = deque([root])
    while queue:
        current = queue.popleft()
        if current.is_dir():
            for child in current.iterdir():
                queue.append(child)
        else:
            yield current


def is_cxx_source_file(path: Path) -> bool:
    return path.exists() and path.suffix in (
        ".cc",
        ".cp",
        ".cxx",
        ".cpp",
        ".CPP",
        ".c++",
        ".C",
        ".c",
    )


def is_cmake_file(path: Path) -> bool:
    return path.exists() and path.name.startswith("CMakeLists.txt")


def is_catch2_file(path: Path) -> bool:
    if not is_cxx_source_file(path):
        return False
    try:
        content = open(path).read()
        if (
            "#define CATCH_CONFIG_RUNNER" in content
            or "#define CATCH_CONFIG_MAIN" in content
        ):
            return True
        return False
    except UnicodeDecodeError:
        return False


def is_makefile_file(path: Path) -> bool:
    return path.exists() and path.name == "Makefile"


def is_makefile_target(makefile: Path, target: str) -> bool:
    if not is_makefile_file(makefile):
        return False
    proc = subprocess.run(
        ["make", "-qp"], cwd=makefile.parent, capture_output=True, text=True
    )
    for line in proc.stdout.splitlines():
        if line.strip().startswith(f"{target}:"):
            return True
    return False


def is_valid_target(target: str) -> bool:
    if len(target) < 3:
        return False
    if target in (
        "all",
        "install",
        "depend",
        "package",
        "test",
        "package_source",
        "edit_cache",
        "rebuild_cache",
        "clean",
        "help",
        "build.ninja",
        "ALL_BUILD",
        "ZERO_CHECK",
        "INSTALL",
        "RUN_TESTS",
        "PACKAGE",
    ):
        return False
    if "catch2" in target.lower():
        return False
    for banned in ("experimental", "nightly", "continuous", "cache", "cmake"):
        if banned in target.lower():
            return False
    if target.endswith(".obj") or target.endswith(".i") or target.endswith(".s"):
        return False
    return True



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...formatters.dispatcher import FormatPackage


class PenaltyInterface(ABC):
    @abstractmethod
    def get_penalty(self) -> float:
        pass

    @abstractmethod
    def get_output(self) -> FormatPackage:
        pass



from pathlib import Path
from typing import List, Protocol, TypedDict, Union, runtime_checkable

try:
    from typing import NotRequired
except ImportError:
    from typing_extensions import NotRequired


@runtime_checkable
class TestCaseProtocol(Protocol):
    __test__: bool = False

    def get_name(self) -> str: ...
    def get_input(self) -> str: ...
    def get_expected_output(self) -> str: ...


@runtime_checkable
class FlaggedTestCaseProtocol(TestCaseProtocol, Protocol):
    def get_flags(self) -> List[str | Path]: ...


@runtime_checkable
class WeightedTestCaseProtocol(TestCaseProtocol, Protocol):
    def get_weight(self) -> float: ...


@runtime_checkable
class FlaggedWeightedTestCaseProtocol(
    FlaggedTestCaseProtocol, WeightedTestCaseProtocol, Protocol
):
    pass


class TestCaseDict(TypedDict):
    __test__ = False  # type: ignore

    name: str
    input: str
    expected_output: str
    flags: NotRequired[List[str | Path]]
    weight: NotRequired[float]


TestCase = Union[
    TestCaseProtocol,
    FlaggedTestCaseProtocol,
    WeightedTestCaseProtocol,
    FlaggedWeightedTestCaseProtocol,
    TestCaseDict,
]


class TemplateSubstitution:
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs


TSub = TemplateSubstitution  # faster-to-type alias.



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Sequence

from .simple import make_tests_from_strs
from .validation import validate_unique_argument

if TYPE_CHECKING:
    from ...builders.interfaces.builder import BuilderInterface
    from .output_tests import CLIOutputTest


def make_tests_from_files(
    *,  # kwargs-only; to avoid confusion with argument sequence.
    builder: BuilderInterface,
    names: Sequence[str],
    input_files: Optional[Sequence[Path]] = None,
    input_strs: Optional[Sequence[str]] = None,
    expected_output_files: Optional[Sequence[Path]] = None,
    expected_output_strs: Optional[Sequence[str]] = None,
    weights: Optional[Sequence[float]] = None,  # Defaults to equal-weight.
    flag_sets: Optional[Sequence[List[str]]] = None,
) -> List[CLIOutputTest]:

    # Ensure only one input is specified, then convert files to strs.
    validate_unique_argument(input_files=input_files, input_strs=input_strs)
    if input_strs is None:
        input_strs = []
        assert input_files is not None
        for input_file in input_files:
            with open(Path(input_file), "r") as f:
                input_strs.append(f.read())

    # Ensure only one expected output is specified, then convert files to strs.
    validate_unique_argument(
        expected_output_files=expected_output_files,
        expected_output_strs=expected_output_strs,
    )
    if expected_output_strs is None:
        expected_output_strs = []
        assert expected_output_files is not None
        for expected_output_file in expected_output_files:
            with open(Path(expected_output_file), "r") as f:
                expected_output_strs.append(f.read())

    return make_tests_from_strs(
        builder=builder,
        names=names,
        inputs=input_strs,
        expected_outputs=expected_output_strs,
        weights=weights,
        flag_sets=flag_sets,
    )



import difflib
from abc import ABC, abstractmethod
from datetime import timedelta
from typing import Any, Dict, Generic, List, Mapping, Tuple, TypeVar, cast

from colorama import Back, Fore

from ...types import (
    AssignmentMetadataOutput,
    ByteStreamComparisonOutput,
    CommandOutput,
    StreamOutput,
    UnitTestCase,
    UnitTestSuite,
    ValgrindOutput,
    is_successful_test,
)
from .dispatcher import register_format

T = TypeVar("T", bound=Mapping[str, Any])


class FormatterInterface(ABC, Generic[T]):

    @classmethod
    @abstractmethod
    def to_string(cls, data: T) -> str:
        pass


class StreamOutputInterface(FormatterInterface[StreamOutput]):
    _prefix: str = ""
    _suffix: str = ""
    _empty: str = ""

    def __init_subclass__(cls, *, prefix: str, suffix: str, empty: str):
        cls._prefix = prefix
        cls._suffix = suffix
        cls._empty = empty

    @classmethod
    def to_string(cls, data: StreamOutput) -> str:
        return (
            f"{cls.get_prefix()}{data['stream_contents']}{cls.get_suffix()}"
            if data["stream_contents"]
            else cls.get_empty()
        )

    @classmethod
    def get_prefix(cls):
        return cls._prefix

    @classmethod
    def get_suffix(cls):
        return cls._suffix

    @classmethod
    def get_empty(cls):
        return cls._empty


@register_format("raw")
class RawFormatter(FormatterInterface[StreamOutput]):
    @classmethod
    def to_string(cls, data: StreamOutput) -> str:
        return data["stream_contents"]


@register_format("stdin")
class STDInFormatter(
    StreamOutputInterface,
    prefix=f"<{Fore.MAGENTA}BEGIN STDIN{Fore.RESET}>\n",
    suffix=f"\n<{Fore.MAGENTA}END STDIN{Fore.RESET}>",
    empty=f"<{Fore.MAGENTA}EMPTY STDIN{Fore.RESET}>",
): ...


@register_format("stdout")
class STDOutFormatter(
    StreamOutputInterface,
    prefix=f"<{Fore.BLUE}BEGIN STDOUT{Fore.RESET}>\n",
    suffix=f"\n<{Fore.BLUE}END STDOUT{Fore.RESET}>",
    empty=f"<{Fore.BLUE}EMPTY STDOUT{Fore.RESET}>",
): ...


@register_format("stderr")
class STDErrFormatter(
    StreamOutputInterface,
    prefix=f"<{Fore.RED}BEGIN STDERR{Fore.RESET}>\n",
    suffix=f"\n<{Fore.RED}END STDERR{Fore.RESET}>",
    empty=f"<{Fore.RED}EMPTY STDERR{Fore.RESET}>",
): ...


@register_format("expected-stdout")
class ExpectedSTDOutFormatter(
    StreamOutputInterface,
    prefix=f"<{Fore.BLUE}BEGIN EXPECTED STDOUT{Fore.RESET}>\n",
    suffix=f"\n<{Fore.BLUE}END EXPECTED STDOUT{Fore.RESET}>",
    empty=f"<{Fore.BLUE}EMPTY EXPECTED STDOUT{Fore.RESET}>",
): ...


@register_format("actual-stdout")
class ActualSTDOutFormatter(
    StreamOutputInterface,
    prefix=f"<{Fore.LIGHTMAGENTA_EX}BEGIN ACTUAL STDOUT{Fore.RESET}>\n",
    suffix=f"\n<{Fore.LIGHTMAGENTA_EX}END ACTUAL STDOUT{Fore.RESET}>",
    empty=f"<{Fore.LIGHTMAGENTA_EX}EMPTY ACTUAL STDOUT{Fore.RESET}>",
): ...


@register_format("byte-cmp")
class ByteCmpFormatter(FormatterInterface[ByteStreamComparisonOutput]):
    @staticmethod
    def bytes_to_safe_ascii(data: bytes) -> str:
        return "".join(
            (
                chr(b)
                if 32 <= b <= 126
                else (
                    r"\n"
                    if b == ord("\n")
                    else (
                        r"\t"
                        if b == ord("\t")
                        else r"\r" if b == ord("\r") else f"\\x{b:02x}"
                    )
                )
            )
            for b in data
        )

    @staticmethod
    def abbreviate(a: bytes, max_len: int = 1024) -> str:
        if len(a) <= max_len:
            return ByteCmpFormatter.bytes_to_safe_ascii(a)
        return (
            f"{ByteCmpFormatter.bytes_to_safe_ascii(a[:max_len//2])}"
            " (...) "
            f"{ByteCmpFormatter.bytes_to_safe_ascii(a[-max_len//2:])}"
        )

    @staticmethod
    def static_comparison(expected: bytes, actual: bytes, max_len: int = 32) -> str:
        differences: List[Tuple[int, int, int]] = []
        for i, (b_a, b_b) in enumerate(zip(expected, actual)):
            if len(differences) > max_len:
                break
            if b_a != b_b:
                differences.append((b_a, b_b, i))

        if len(differences) == 0:
            return (
                f"  {Fore.CYAN}*{Fore.RESET} ({Fore.GREEN}No differences!{Fore.RESET})"
            )
        elif len(differences) <= max_len:
            return "\n".join(
                f"  {Fore.CYAN}*{Fore.RESET} Expected: ({Back.GREEN}{b_a:02x}{Back.RESET}); Actual: ({Back.RED}{b_b:02x}{Back.RESET}); @ Location: 0x{b_l:08x}/{b_l}"
                for b_a, b_b, b_l in differences
            )
        return "\n".join(
            [
                f"  {Fore.CYAN}*{Fore.RESET} Expected: ({Back.GREEN}{b_a:02x}{Back.RESET}); Actual: ({Back.RED}{b_b:02x}{Back.RESET}); @ Location: 0x{b_l:08x}/{b_l}"
                for b_a, b_b, b_l in differences[:max_len]
            ]
            + [f"  {Fore.CYAN}*{Fore.RESET} ..."]
        )

    @staticmethod
    def string_comparison(expected: bytes, actual: bytes, max_len: int = 32768) -> str:
        if len(expected) > max_len or len(actual) > max_len:
            return f"  {Fore.CYAN}*{Fore.RESET} ({Fore.RED}Too long to compare!{Fore.RESET})"
        str_a = ByteCmpFormatter.bytes_to_safe_ascii(expected)
        str_b = ByteCmpFormatter.bytes_to_safe_ascii(actual)
        diff = difflib.ndiff(str_a, str_b)

        out_a: List[str] = []
        out_b: List[str] = []

        for d in diff:
            if d.startswith("- "):
                out_a.append(f"{Back.GREEN}{d[2:]}{Back.RESET}")
            elif d.startswith("+ "):
                out_b.append(f"{Back.RED}{d[2:]}{Back.RESET}")
            elif d.startswith("? "):
                continue
            else:
                out_a.append(d[2:])
                out_b.append(d[2:])

        return (
            f"  {Fore.CYAN}*{Fore.RESET} Expected (Missing Highlighted): "
            + "".join(out_a)
            + f"\n  {Fore.CYAN}*{Fore.RESET} Actual (Insertions Highlighted): "
            + "".join(out_b)
        )

    @classmethod
    def to_string(cls, data: ByteStreamComparisonOutput):
        expected: bytes = data["stream_expected_bytes"]
        actual: bytes = data["stream_actual_bytes"]

        len_expected_output = f"EXPECTED OUTPUT LENGTH: {len(expected)}"
        len_actual_output = f"ACTUAL OUTPUT LENGTH: {len(actual)}"

        raw_expected_output: str = (
            "RAW EXPECTED OUTPUT: " + ByteCmpFormatter.abbreviate(expected)
        )
        raw_actual_output: str = "RAW ACTUAL OUTPUT: " + ByteCmpFormatter.abbreviate(
            actual
        )

        chr_output_comparison: str = (
            "BYTE COMPARISON: \n" + ByteCmpFormatter.static_comparison(expected, actual)
        )
        str_output_comparison: str = (
            "STRING COMPARISON: \n"
            + ByteCmpFormatter.string_comparison(expected, actual)
        )

        return (
            f"<{Fore.LIGHTYELLOW_EX}BEGIN BYTE COMPARISON{Fore.RESET}>\n"
            f"{len_expected_output}\n"
            f"{len_actual_output}\n\n"
            f"{raw_expected_output}\n"
            f"{raw_actual_output}\n\n"
            f"{chr_output_comparison}\n\n"
            f"{str_output_comparison}\n"
            f"<{Fore.LIGHTYELLOW_EX}END BYTE COMPARISON{Fore.RESET}>"
        )


@register_format("unit-tests")
class UnitTestFormatter(FormatterInterface[UnitTestSuite | UnitTestCase]):
    @classmethod
    def to_string(cls, data: UnitTestSuite | UnitTestCase):
        strings: List[str] = []
        color: str
        if "success" in data.keys():
            data = cast(UnitTestCase, data)
            success = is_successful_test(data)
            color = Fore.GREEN if success else Fore.RED
            if not success:
                strings.append(
                    f"<{color}BEGIN FAILED TEST CASE '{data['name']}'{Fore.RESET}>"
                )
                strings.append(
                    "\n".join(f"  {line}" for line in data["output"].split("\n"))
                )
                strings.append(
                    f"<{color}END FAILED TEST CASE '{data['name']}'{Fore.RESET}>\n"
                )
            else:
                strings.append(
                    f"<{color}PASSED TEST CASE '{data['name']}'{Fore.RESET}>"
                )
        else:
            data = cast(UnitTestSuite, data)
            color = Fore.GREEN if is_successful_test(data) else Fore.RED
            desc = "PASSED" if is_successful_test(data) else "FAILED"
            rec_string: str = "\n".join(cls.to_string(_case) for _case in data["cases"])

            strings.append(
                f"<{color}BEGIN {desc} TEST SUITE '{data['name']}'{Fore.RESET}>"
            )
            strings.append("\n".join(f"  {line}" for line in rec_string.split("\n")))
            strings.append(
                f"<{color}END {desc} TEST SUITE '{data['name']}'{Fore.RESET}>\n"
            )

        return "\n".join(strings)


@register_format("build-fail")
class BuildFailureFormatter(FormatterInterface[Dict]):
    @classmethod
    def to_string(cls, data: Dict):
        return f"{Fore.RED}<NO EXECUTABLE GENERATED>{Fore.RESET}"


@register_format("valgrind")
class ValgrindFormatter(FormatterInterface[ValgrindOutput]):
    @classmethod
    def to_string(cls, data: ValgrindOutput):
        leak_summary = data["leaks"]
        warning_summary = data["warnings"]

        def_lost_color = (
            Fore.LIGHTGREEN_EX if leak_summary.definitely_lost.is_safe else Fore.RED
        )
        ind_lost_color = (
            Fore.LIGHTGREEN_EX if leak_summary.indirectly_lost.is_safe else Fore.RED
        )
        pos_lost_color = (
            Fore.LIGHTGREEN_EX if leak_summary.possibly_lost.is_safe else Fore.RED
        )
        leak_text = (
            f"{Fore.LIGHTGREEN_EX}VALGRIND LEAK SUMMARY{Fore.RESET}:\n"  # I know these pluses aren't necessary, but I'm doing it for back-compatibility because linters yell at me for line-continuation.
            + f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} {def_lost_color}{leak_summary.definitely_lost.bytes}{Fore.RESET} bytes, {def_lost_color}{leak_summary.definitely_lost.blocks}{Fore.RESET} blocks {def_lost_color}definitely lost{Fore.RESET}.\n"
            + f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} {ind_lost_color}{leak_summary.indirectly_lost.bytes}{Fore.RESET} bytes, {ind_lost_color}{leak_summary.indirectly_lost.blocks}{Fore.RESET} blocks {ind_lost_color}indirectly lost{Fore.RESET}.\n"
            + f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} {pos_lost_color}{leak_summary.possibly_lost.bytes}{Fore.RESET} bytes, {pos_lost_color}{leak_summary.possibly_lost.blocks}{Fore.RESET} blocks {pos_lost_color}possibly lost{Fore.RESET}.\n"
            + f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} {Fore.LIGHTGREEN_EX}{leak_summary.still_reachable.bytes}{Fore.RESET} bytes, {Fore.LIGHTGREEN_EX}{leak_summary.still_reachable.blocks}{Fore.RESET} blocks {Fore.LIGHTGREEN_EX}still reachable{Fore.RESET}."
        )

        warning = warning_summary.model_dump()
        output = [f"{Fore.LIGHTGREEN_EX}VALGRIND WARNING SUMMARY{Fore.RESET}:"]
        output += [
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} {Fore.LIGHTGREEN_EX if v == 0 else Fore.RED}{v}{Fore.RESET} `{Fore.LIGHTRED_EX}{k.replace('_', ' ').upper()}{Fore.RESET}` warnings encountered."
            for k, v in warning.items()
        ]

        warning_text = "\n".join(output)

        return f"{leak_text}\n\n{warning_text}"


@register_format("command")
class CommandFormatter(FormatterInterface[CommandOutput]):
    @classmethod
    def to_string(cls, data: CommandOutput):
        return f"Ran `{Fore.MAGENTA}{data['command']}{Fore.RESET}` in CLI with exit code, \"{data['exit_code']}\"."


@register_format("assignment-metadata")
class MetadataFormatter(FormatterInterface[AssignmentMetadataOutput]):
    @staticmethod
    def format_timedelta(td: timedelta) -> str:
        if td < timedelta(0):
            return "0s"
        total_seconds = int(td.total_seconds())
        milliseconds = int(td.microseconds / 1000)

        weeks, remainder = divmod(total_seconds, 7 * 24 * 3600)
        days, remainder = divmod(remainder, 24 * 3600)
        hours, remainder = divmod(remainder, 3600)
        minutes, seconds = divmod(remainder, 60)

        parts = []
        if weeks:
            parts.append(f"{weeks}w")
        if days:
            parts.append(f"{days}d")
        if hours:
            parts.append(f"{hours}h")
        if minutes:
            parts.append(f"{minutes}m")
        if seconds:
            parts.append(f"{seconds}s")
        if milliseconds:
            parts.append(f"{milliseconds}ms")

        return " ".join(parts) if parts else "0s"

    @classmethod
    def to_string(cls, data: AssignmentMetadataOutput) -> str:
        return (
            f"`{Fore.MAGENTA}{data['metadata'].assignment_name}{Fore.RESET}` made by {Fore.MAGENTA}{f'{Fore.RESET}, {Fore.MAGENTA}'.join(data['metadata'].assignment_authors)}{Fore.RESET}.\n"
            + f'Submission date: {Fore.MAGENTA}{data["metadata"].assignment_submit_date.strftime("%Y-%m-%d %H:%M:%S.%f")}{Fore.RESET}.\n'
            + f'Due date: {Fore.MAGENTA}{data["metadata"].assignment_due_date.strftime("%Y-%m-%d %H:%M:%S.%f")}{Fore.RESET}.\n'
            + f"Time left: {Fore.MAGENTA}{cls.format_timedelta(data['metadata'].assignment_due_date - data['metadata'].assignment_submit_date)}{Fore.RESET}.\n\n"
            + f"Assignment graded with `{Fore.GREEN}{data['metadata'].library_name}{Fore.RESET}` (version {Fore.GREEN}{data['metadata'].library_version}{Fore.RESET}), made by {Fore.GREEN}{f'{Fore.RESET}, {Fore.GREEN}'.join(data['metadata'].library_authors)}.{Fore.RESET}\n\n"
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from ....types import Command


class BuilderInterface(ABC):
    def __init__(self):
        super().__init__()
        self._project_root: Optional[Path] = None
        self._build_error: bool = False

    def set_build_error(self, build_error: bool = True) -> None:
        self._build_error = build_error

    def get_build_error(self) -> bool:
        return self._build_error

    def set_project_root(self, path: Path) -> None:
        self._project_root = path

    def get_project_root(self) -> Path:
        assert self._project_root is not None
        return self._project_root

    def wrap_args(self) -> bool:
        return False

    @abstractmethod
    def build_project(self) -> None:
        pass

    @abstractmethod
    def get_build_directory(self) -> Path:
        pass

    @abstractmethod
    def get_start_command(self) -> Command:
        pass



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple

from ....data.messages import MessageConfig
from ....os.cmd import run_cmd
from ..interfaces.cli_test import CLITest
from ..interfaces.output_test import OutputTestInterface

if TYPE_CHECKING:
    from ....types import Command
    from ...addons.addon import ExecAddonInterface
    from ...builders.interfaces.builder import BuilderInterface


class CLIOutputTest(CLITest, OutputTestInterface):
    def __init__(self):
        super().__init__()
        self._name: Optional[str] = None
        self._builder: Optional[BuilderInterface] = None

        self._expected_stdout: Optional[str] = None
        self._actual_stdout: Optional[str] = None
        self._stderr: Optional[str] = None

    @classmethod
    def make(
        cls,
        name: str,
        builder: BuilderInterface,
        stdin: str,
        expected_stdout: str,
        weight: float = 1.0,
        args: Optional[Command] = None,
        working_dir: Optional[Path] = None,
        wrap_args: bool = False,
    ) -> CLIOutputTest:
        test = cls()

        test.set_name(name)
        test.set_builder(builder)
        test.set_input(stdin)
        test.set_expected_stdout(expected_stdout)
        test.set_weight(weight)

        test.set_wrap_args(wrap_args)
        if working_dir is not None:
            test.set_working_dir(working_dir)
        if args is not None:
            test.set_args(args)

        return test

    def set_builder(self, builder: BuilderInterface) -> None:
        self._builder = builder

    def get_builder(self) -> BuilderInterface:
        assert self._builder is not None
        return self._builder

    def set_name(self, name: str):
        self._name = name

    def _run_test(self) -> Tuple[int, Command]:
        builder = self.get_builder()
        builder.build_project()
        if builder.get_build_error():
            self.force_fail()
            self.add_to_output("build-fail", {})
            return 1, []

        self.set_working_dir(builder.get_build_directory())

        args = self.get_args()
        command = builder.get_start_command()
        if not command:
            self.force_fail()
            self.get_builder().set_build_error(True)
            self.add_to_output("build-fail", {})
            return 1, []
        command += args

        _tmp_stdout: List[str] = []
        _tmp_stderr: List[str] = []
        try:
            result = run_cmd(
                command,
                self.get_input(),
                [],
                _tmp_stdout,
                _tmp_stderr,
                self.get_working_dir(),
            )
        except FileNotFoundError:
            self.force_fail()
            self.get_builder().set_build_error(True)
            self.add_to_output("build-fail", {})
            return 1, []

        self._actual_stdout = _tmp_stdout.pop()
        self._stderr = _tmp_stderr.pop()

        return result.returncode, command

    def set_expected_stdout(self, expected_stdout: str):
        self._expected_stdout = expected_stdout

    def get_expected_output(self) -> str:
        if self.get_builder().get_build_error():
            return MessageConfig.DEFAULT_BUILD_ERROR
        assert self._expected_stdout is not None
        return self._expected_stdout

    def get_actual_output(self) -> str:
        if self._actual_stdout is None:
            _, _ = self._run_test()
        if self.get_builder().get_build_error():
            return MessageConfig.DEFAULT_BUILD_ERROR
        assert self._actual_stdout is not None
        return self._actual_stdout

    def get_error(self) -> str:
        if self._stderr is None:
            _, _ = self._run_test()
        if self.get_builder().get_build_error():
            return MessageConfig.DEFAULT_BUILD_ERROR
        assert self._stderr is not None
        return self._stderr

    def get_score(self) -> float:
        if self.get_builder().get_build_error():
            if not self._run:
                self.add_to_output("build-fail", {})
                self._run = True
            return 0.0
        return OutputTestInterface.get_score(self)

    def get_name(self) -> str:
        assert self._name is not None
        return self._name

    def add_exec_addon(self, addon: ExecAddonInterface):
        addon.set_builder(self.get_builder())
        addon.set_args(self.get_args())
        addon.set_input(self.get_input())
        self.add_addon(addon)



from __future__ import annotations

import difflib
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from .interfaces.test import TestInterface

if TYPE_CHECKING:
    from ...types import ByteStreamComparisonOutput


class FileTest(TestInterface):
    def __init__(self):
        super().__init__()
        self._name: str = ""
        self._base_filename: Optional[Path] = None
        self._target_filename: Optional[Path] = None
        self._run: bool = False

    @classmethod
    def make(
        cls, name: str, expected_file: Path, actual_file: Path, visible: bool = True
    ) -> FileTest:
        test = cls()
        test.set_name(name)
        test.set_base_filename(expected_file)
        test.set_target_filename(actual_file)
        test.set_visibility(visible)
        return test

    def set_name(self, name: str):
        self._name = name

    def get_name(self) -> str:
        return self._name

    def set_base_filename(self, path: Path):
        self._base_filename = path

    def get_base_filename(self):
        assert self._base_filename is not None
        return self._base_filename

    def set_target_filename(self, path: Path):
        self._target_filename = path

    def get_target_filename(self):
        assert self._target_filename is not None
        return self._target_filename

    def get_expected_output(self) -> bytes:
        with open(self.get_base_filename(), "rb") as f:
            return f.read()

    def get_actual_output(self):
        with open(self.get_target_filename(), "rb") as f:
            return f.read()

    def get_score(self) -> float:
        streams: ByteStreamComparisonOutput = {
            "stream_expected_bytes": self.get_expected_output(),
            "stream_actual_bytes": self.get_actual_output(),
        }
        if not self._run:  # stop duplicate appending
            self.add_to_output("byte-cmp", streams)
            self._run = True

        return (
            (
                difflib.SequenceMatcher(
                    None,
                    streams["stream_actual_bytes"],
                    streams["stream_expected_bytes"],
                ).ratio()
            )
            * self.get_max_score()
            * self.get_weight()
        )



from typing import List


class CxxConfig:
    DEFAULT_CXX_STANDARD: str = "c++20"
    DEFAULT_CXX_COMPILATION_FLAGS: List[str] = [
        "-Wall",
        "-Wextra",
        "-Wpedantic",
        "-Wshadow",
        "-Wconversion",
        "-Wsign-conversion",
        "-Wnull-dereference",
    ]
    DEFAULT_CMAKE_COMPILATION_FLAGS: List[str] = []



import random
import string
import sys
from pathlib import Path

from .data.paths import PathConfig


def random_name(length: int = 20) -> str:
    return "".join(
        random.choices(
            string.ascii_lowercase + string.ascii_uppercase + string.digits, k=length
        )
    )


def random_working_directory() -> Path:
    directory = PathConfig.DEFAULT_ROOT_PATH / random_name()
    directory.mkdir(parents=True, exist_ok=False)
    return directory


def random_executable() -> str:
    return random_name() + ".exe" if sys.platform.startswith("win") else random_name()



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Generator, List

from .output_tests import CLIOutputTest
from .types import (
    FlaggedTestCaseProtocol,
    FlaggedWeightedTestCaseProtocol,
    TestCase,
    TestCaseProtocol,
    WeightedTestCaseProtocol,
)

if TYPE_CHECKING:
    from ...builders.interfaces.builder import BuilderInterface


def make_tests_from_generator(builder: BuilderInterface):
    def decorator(
        generator: Callable[[], Generator[TestCase, None, None]],
    ) -> Callable[[], Generator[TestCase, None, None]]:
        generated_tests: List[CLIOutputTest] = []
        for test_case in generator():
            if isinstance(test_case, WeightedTestCaseProtocol):
                test = CLIOutputTest.make(
                    name=test_case.get_name(),
                    builder=builder,
                    stdin=test_case.get_input(),
                    expected_stdout=test_case.get_expected_output(),
                    weight=test_case.get_weight(),
                    args=[],
                    working_dir=builder.get_build_directory(),
                    wrap_args=builder.wrap_args(),
                )
                generated_tests.append(test)
            elif isinstance(test_case, TestCaseProtocol):
                test = CLIOutputTest.make(
                    name=test_case.get_name(),
                    builder=builder,
                    stdin=test_case.get_input(),
                    expected_stdout=test_case.get_expected_output(),
                    weight=1.0,
                    args=[],
                    working_dir=builder.get_build_directory(),
                    wrap_args=builder.wrap_args(),
                )
                generated_tests.append(test)
            elif isinstance(test_case, FlaggedWeightedTestCaseProtocol):
                test = CLIOutputTest.make(
                    name=test_case.get_name(),
                    builder=builder,
                    stdin=test_case.get_input(),
                    expected_stdout=test_case.get_expected_output(),
                    weight=test_case.get_weight(),
                    args=test_case.get_flags(),
                    working_dir=builder.get_build_directory(),
                    wrap_args=builder.wrap_args(),
                )
                generated_tests.append(test)
            elif isinstance(test_case, FlaggedTestCaseProtocol):
                test = CLIOutputTest.make(
                    name=test_case.get_name(),
                    builder=builder,
                    stdin=test_case.get_input(),
                    expected_stdout=test_case.get_expected_output(),
                    weight=1.0,
                    args=test_case.get_flags(),
                    working_dir=builder.get_build_directory(),
                    wrap_args=builder.wrap_args(),
                )
                generated_tests.append(test)
            elif isinstance(test_case, dict):
                if "weight" in test_case:
                    weight = test_case["weight"]
                else:
                    weight = 1.0
                if "flags" in test_case:
                    flags = test_case["flags"]
                else:
                    flags = []

                test = CLIOutputTest.make(
                    name=test_case["name"],
                    builder=builder,
                    stdin=test_case["input"],
                    expected_stdout=test_case["expected_output"],
                    weight=weight,
                    args=flags,
                    working_dir=builder.get_build_directory(),
                    wrap_args=builder.wrap_args(),
                )
                generated_tests.append(test)
            else:
                raise ValueError(
                    f"`generator` passed to `make_tests_from_generator` produced a type, `{type(test_case)}`, which is not supported."
                )
        return generator

    return decorator



from __future__ import annotations

import os
import re
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

from pydantic import BaseModel, Field

from ...data.build import BuildConfig
from ...data.penalties import PenaltyConfig
from ...random_utils import random_name
from ..formatters.dispatcher import FormatPackage
from .addon import ExecAddonInterface

if TYPE_CHECKING:
    from ...types import ValgrindOutput as ValgrindOutput_t


class LossEntry(BaseModel):
    bytes: int = Field(default=0)
    blocks: int = Field(default=0)

    @property
    def is_safe(self) -> bool:
        return not self.bytes and not self.blocks


class ValgrindLeakSummary(BaseModel):
    definitely_lost: LossEntry = Field(default_factory=LossEntry)
    indirectly_lost: LossEntry = Field(default_factory=LossEntry)
    possibly_lost: LossEntry = Field(default_factory=LossEntry)
    still_reachable: LossEntry = Field(default_factory=LossEntry)

    @property
    def is_safe(self) -> bool:
        return (
            self.definitely_lost.is_safe
            and self.indirectly_lost.is_safe
            and self.possibly_lost.is_safe
        )


class ValgrindWarningSummary(BaseModel):
    invalid_read: int = Field(default=0)
    invalid_write: int = Field(default=0)
    invalid_free: int = Field(default=0)
    mismatched_free: int = Field(default=0)
    uninitialized_value: int = Field(default=0)
    conditional_jump: int = Field(default=0)
    syscall_param: int = Field(default=0)
    overlap: int = Field(default=0)
    other: int = Field(default=0)  # fallback bucket

    @property
    def is_safe(self) -> bool:
        return (
            not self.invalid_read
            and not self.invalid_write
            and not self.invalid_free
            and not self.mismatched_free
            and not self.uninitialized_value
            and not self.conditional_jump
            and not self.syscall_param
            and not self.overlap
            and not self.other
        )


class ValgrindOutput:
    LEAK_REGEX = re.compile(
        r"(\d+)\s+bytes in\s+(\d+)\s+blocks are (definitely lost|indirectly lost|possibly lost|still reachable)"
    )
    WARNING_PATTERNS = {
        "invalid_read": re.compile(r"Invalid read"),
        "invalid_write": re.compile(r"Invalid write"),
        "invalid_free": re.compile(r"Invalid free"),
        "mismatched_free": re.compile(r"Mismatched free"),
        "uninitialized_value": re.compile(r"uninitialised value"),
        "conditional_jump": re.compile(
            r"Conditional jump or move depends on uninitialised value"
        ),
        "syscall_param": re.compile(r"Syscall param"),
        "overlap": re.compile(r"overlap"),
    }

    def __init__(self, stderr: str):
        self._stderr: str = stderr
        self._warnings: ValgrindWarningSummary = ValgrindWarningSummary()
        self._leaks: ValgrindLeakSummary = ValgrindLeakSummary()
        self.parse_stderr()

    @classmethod
    def parse_valgrind_log(
        cls, stderr: str
    ) -> tuple[ValgrindLeakSummary, ValgrindWarningSummary]:
        # Init structures
        leaks: ValgrindLeakSummary = ValgrindLeakSummary()
        warnings: ValgrindWarningSummary = ValgrindWarningSummary()
        warnings.other = 0

        for line in stderr.split("\n"):
            # --- Leak parsing ---
            leak_match = cls.LEAK_REGEX.search(line)
            if leak_match:
                bytes_count = int(leak_match.group(1).replace(",", ""))
                blocks_count = int(leak_match.group(2).replace(",", ""))
                kind = leak_match.group(3).replace(" ", "_")  # normalize to dict key
                loss_entry = getattr(leaks, kind)
                prev_bytes = loss_entry.bytes
                prev_blocks = loss_entry.blocks
                setattr(
                    leaks,
                    kind,
                    LossEntry(
                        bytes=prev_bytes + bytes_count,
                        blocks=prev_blocks + blocks_count,
                    ),
                )
                continue

            # --- Warning parsing ---
            matched = False
            for key, pattern in cls.WARNING_PATTERNS.items():
                if pattern.search(line):
                    setattr(warnings, key, getattr(warnings, key) + 1)
                    matched = True
                    break
            if not matched and "==" in line and "==" in line.strip():
                # heuristic: unknown Valgrind warning line
                if not any(x in line for x in ["lost", "reachable"]):
                    warnings.other += 1

        return leaks, warnings

    def parse_stderr(self):
        self._leaks, self._warnings = self.parse_valgrind_log(self._stderr)

    def get_leaks(self) -> ValgrindLeakSummary:
        return self._leaks

    def get_warnings(self) -> ValgrindWarningSummary:
        return self._warnings


def valgrind(
    cmd: List[str | Path], stdin: Optional[str] = None
) -> tuple[ValgrindLeakSummary, ValgrindWarningSummary]:

    if sys.platform.startswith("win"):
        return ValgrindLeakSummary(), ValgrindWarningSummary()

    valgrind_file = f"valgrind-{random_name()}.log"
    with open(os.devnull, "w") as devnull:
        result = subprocess.run(
            [
                "valgrind",
                "--leak-check=full",
                "--show-leak-kinds=all",
                f"--log-file={valgrind_file}",
            ]
            + cmd,
            input=stdin,
            stdout=devnull,
            stderr=devnull,
            text=True,
            timeout=BuildConfig.DEFAULT_EXECUTABLE_TIMEOUT,
        )

    if result.returncode != 0:
        if Path(valgrind_file).is_file():
            os.remove(valgrind_file)
        return ValgrindLeakSummary(), ValgrindWarningSummary()

    with open(valgrind_file, "r", encoding="utf-8", errors="ignore") as f:
        valgrind_log = f.read()
    valgrind_output = ValgrindOutput(valgrind_log)
    os.remove(valgrind_file)

    return valgrind_output.get_leaks(), valgrind_output.get_warnings()


class ValgrindAddon(ExecAddonInterface):
    def __init__(self):
        super().__init__()
        self._output: Optional[ValgrindLeakSummary] = None
        self._warnings: Optional[ValgrindWarningSummary] = None

    def run(self) -> None:
        self._output, self._warnings = valgrind(
            self.get_builder().get_start_command() + self.get_args(), self.get_input()
        )

    def get_penalty(self) -> float:
        if self._output is None:
            self.run()
        assert self._output is not None
        return (
            PenaltyConfig.DEFAULT_VALGRIND_LEAK_PENALTY
            if not self._output.is_safe
            else 1.0
        )

    def get_output(self) -> FormatPackage:
        if self._output is None or self._warnings is None:
            self.run()
        assert self._output is not None
        assert self._warnings is not None

        data: ValgrindOutput_t = {"leaks": self._output, "warnings": self._warnings}

        return FormatPackage(
            label="valgrind",
            data=data,
        )



import shutil
import weakref
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from ...random_utils import random_working_directory
from ..builders.interfaces.builder import BuilderInterface


class UnitTesterInterface(BuilderInterface, ABC):
    def __init__(self):
        super().__init__()
        self._testing_root: Optional[Path] = None
        self._instance_root: Path = random_working_directory()

        weakref.finalize(self, self._cleanup, self.get_instance_root())

    def set_testing_root(self, path: Path):
        self._testing_root = path

    def get_testing_root(self):
        assert self._testing_root is not None
        return self._testing_root

    @staticmethod
    def _cleanup(path: Path):
        shutil.rmtree(path, ignore_errors=True)

    @abstractmethod
    def build_test(self) -> None:
        pass

    def set_instance_root(self, path: Path):
        self._instance_root = path

    def get_instance_root(self) -> Path:
        return self._instance_root

    def build_project(self) -> None:
        shutil.copytree(
            self.get_testing_root(), self.get_instance_root(), dirs_exist_ok=True
        )
        shutil.copytree(
            self.get_project_root(), self.get_instance_root(), dirs_exist_ok=True
        )
        self.build_test()



from datetime import datetime, timedelta
from pathlib import Path
from typing import Generator

from testing_utils import create_dummy_submission, get_results

from lograder.grader.builders.dispatcher import ProjectDispatcher
from lograder.grader.submission_handler import SubmissionHandler
from lograder.grader.tests.output_comparison import (
    TestCaseDict,
    TestCaseProtocol,
    make_tests_from_files,
    make_tests_from_generator,
    make_tests_from_strs,
)


def test_cmake_hello_world(tmp_path: Path):
    create_dummy_submission(tmp_path, Path("tests/test-projects/cmake-cmp-project-1"))

    SubmissionHandler.clear()

    builder = ProjectDispatcher()
    builder.set_allowed_project_types(["cmake"])

    make_tests_from_strs(
        builder=builder,
        names=["Test `Hello World`."],
        inputs=[""],
        expected_outputs=["Hello World from `lograder`!"],
    )
    SubmissionHandler.make_submission(
        assignment_name="Hello World from `lograder`!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic compilation process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()

    for test in results["tests"]:
        assert test["score"] == test["max_score"]


def test_cmake_hello_world_bad_1(tmp_path: Path):
    create_dummy_submission(tmp_path, Path("tests/test-projects/cmake-bad-project-1"))

    SubmissionHandler.clear()

    builder = ProjectDispatcher()
    builder.set_allowed_project_types(["cmake"])

    make_tests_from_strs(
        builder=builder,
        names=["Test `Hello World`."],
        inputs=[""],
        expected_outputs=["Hello World from `lograder`!"],
    )
    SubmissionHandler.make_submission(
        assignment_name="Hello World from `lograder`!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic compilation process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()

    for test in results["tests"]:
        assert test["score"] == 0.0


def test_cmake_hello_world_bad_2(tmp_path: Path):
    create_dummy_submission(tmp_path, Path("tests/test-projects/cmake-bad-project-2"))

    SubmissionHandler.clear()

    builder = ProjectDispatcher()
    builder.set_allowed_project_types(["cmake"])

    make_tests_from_strs(
        builder=builder,
        names=["Test `Hello World`."],
        inputs=[""],
        expected_outputs=["Hello World from `lograder`!"],
    )
    SubmissionHandler.make_submission(
        assignment_name="Hello World from `lograder`!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic compilation process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()

    for test in results["tests"]:
        assert test["score"] == 0.0


def test_cmake_echo(tmp_path: Path):
    create_dummy_submission(tmp_path, Path("tests/test-projects/cmake-cmp-project-2"))

    SubmissionHandler.clear()

    builder = ProjectDispatcher()
    builder.set_allowed_project_types(["cmake"])

    make_tests_from_strs(
        builder=builder,
        names=['Echoing "Hello World".'],
        inputs=["Hello World"],
        expected_outputs=["Hello World"],
    )
    SubmissionHandler.make_submission(
        assignment_name="Test `Echo` from `lograder`!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic compilation process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()
    for test in results["tests"]:
        assert test["score"] == test["max_score"]


def test_cmake_echo_file(tmp_path: Path):
    ifile = tmp_path / "input.txt"
    ofile = tmp_path / "output.txt"
    with open(ifile, "w") as f:
        f.write("Hello World")
    with open(ofile, "w") as f:
        f.write("Hello World")

    create_dummy_submission(tmp_path, Path("tests/test-projects/cmake-cmp-project-2"))

    SubmissionHandler.clear()

    builder = ProjectDispatcher()
    builder.set_allowed_project_types(["cmake"])

    make_tests_from_files(
        builder=builder,
        names=['Echoing "Hello World" (both files).'],
        input_files=[ifile],
        expected_output_files=[ofile],
    )
    make_tests_from_files(
        builder=builder,
        names=['Echoing "Hello World" (input file, output string).'],
        input_files=[ifile],
        expected_output_strs=["Hello World"],
    )
    make_tests_from_files(
        builder=builder,
        names=['Echoing "Hello World" (input string, output file).'],
        input_strs=["Hello World"],
        expected_output_files=[ofile],
    )
    make_tests_from_files(
        builder=builder,
        names=['Echoing "Hello World" (both strings).'],
        input_strs=["Hello World"],
        expected_output_strs=["Hello World"],
    )
    SubmissionHandler.make_submission(
        assignment_name="Test `Echo` from `lograder` with files!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic compilation process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()
    for test in results["tests"]:
        assert test["score"] == test["max_score"]


def test_cmake_echo_file_bad(tmp_path: Path):
    ifile = tmp_path / "input.txt"
    ofile = tmp_path / "output.txt"
    with open(ifile, "w") as f:
        f.write("Bye World")
    with open(ofile, "w") as f:
        f.write("Hello World")

    create_dummy_submission(tmp_path, Path("tests/test-projects/cmake-cmp-project-2"))

    SubmissionHandler.clear()

    builder = ProjectDispatcher()
    builder.set_allowed_project_types(["cmake"])

    make_tests_from_files(
        builder=builder,
        names=['Echoing "Hello World" (both files).'],
        input_files=[ifile],
        expected_output_files=[ofile],
    )
    make_tests_from_files(
        builder=builder,
        names=['Echoing "Hello World" (input file, output string).'],
        input_files=[ifile],
        expected_output_strs=["Hello World"],
    )
    make_tests_from_files(
        builder=builder,
        names=['Echoing "Hello World" (input string, output file).'],
        input_strs=["Bye World"],
        expected_output_files=[ofile],
    )
    make_tests_from_files(
        builder=builder,
        names=['Echoing "Hello World" (both strings).'],
        input_strs=["Bye World"],
        expected_output_strs=["Hello World"],
    )
    SubmissionHandler.make_submission(
        assignment_name="Test `Echo` from `lograder` with files!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic compilation process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()
    for test in results["tests"]:
        print(test["output"])
        assert test["score"] == 0


def test_cmake_echo_gen(tmp_path: Path):
    N_TESTS: int = 1

    SubmissionHandler.clear()

    builder = ProjectDispatcher()
    builder.set_allowed_project_types(["cmake"])

    class Case(TestCaseProtocol):
        def __init__(self, num: int):
            self._name: str = str(num)

        def get_name(self):
            return self._name

        def get_input(self):
            return self._name

        def get_expected_output(self):
            return self._name

    @make_tests_from_generator(builder)
    def test_generator_dict() -> Generator[TestCaseDict, None, None]:
        for i in range(N_TESTS):
            yield TestCaseDict(name=f"{i}", input=f"{i}", expected_output=f"{i}")

    @make_tests_from_generator(builder)
    def test_generator_protocol() -> Generator[TestCaseProtocol, None, None]:
        for i in range(N_TESTS):
            yield Case(i)

    create_dummy_submission(tmp_path, Path("tests/test-projects/cmake-cmp-project-2"))

    SubmissionHandler.make_submission(
        assignment_name="Test `Echo` from `lograder` with files!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic compilation process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()

    assert len(results["tests"]) == 2 * N_TESTS

    for test in results["tests"]:
        assert test["score"] == test["max_score"]


def test_cmake_echo_gen_bad(tmp_path: Path):
    N_TESTS: int = 1

    SubmissionHandler.clear()

    builder = ProjectDispatcher()
    builder.set_allowed_project_types(["cmake"])

    class Case(TestCaseProtocol):
        def __init__(self, num: int):
            self._name: str = str(num)

        def get_name(self):
            return self._name

        def get_input(self):
            return self._name + "Buggadoo"

        def get_expected_output(self):
            return self._name

    @make_tests_from_generator(builder)
    def test_generator_dict() -> Generator[TestCaseDict, None, None]:
        for i in range(N_TESTS):
            yield TestCaseDict(
                name=f"{i+100}", input=f"{i+100}", expected_output=f"{i}"
            )

    @make_tests_from_generator(builder)
    def test_generator_protocol() -> Generator[TestCaseProtocol, None, None]:
        for i in range(N_TESTS):
            yield Case(i)

    create_dummy_submission(tmp_path, Path("tests/test-projects/cmake-cmp-project-2"))

    SubmissionHandler.make_submission(
        assignment_name="Test `Echo` from `lograder` with files!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic compilation process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()

    assert len(results["tests"]) == 2 * N_TESTS

    for test in results["tests"]:
        print(test["output"])
        assert test["score"] == 0



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, field_validator, model_validator

from ...types import AscendingOrder, Status, TextFormat, Visibility


# -----------------------------
# JSON schema for a single test case
# -----------------------------
class TestCaseJSON(BaseModel):
    __test__: bool = False  # Prevent pytest from treating this class as a test

    # Core test result data
    score: Optional[float]  # Score for this test
    max_score: Optional[float]  # Maximum possible score
    status: Optional[Status] = None  # Pass/Fail/Other status
    name: Optional[str]  # Name of the test
    name_format: Optional[TextFormat] = "text"  # Format for the test name
    number: Optional[str] = None  # Optional test number/index

    # Output-related fields
    output: Optional[str]  # Output of the test (raw text or ANSI)
    output_format: Optional[TextFormat] = (
        "ansi"  # Format of output (ansi, markdown, etc.)
    )

    # Metadata
    tags: List[str] = Field(default_factory=list)  # Custom tags for filtering
    visibility: Optional[Visibility] = "visible"  # Visibility on Gradescope
    extra_data: Optional[Dict[str, Any]] = Field(  # Arbitrary extra data
        default_factory=dict
    )

    @property
    def is_scored(self) -> bool:
        """Whether this test has been assigned a score."""
        return self.score is not None


# -----------------------------
# JSON schema for leaderboard entries
# -----------------------------
class LeaderboardJSON(BaseModel):
    name: str  # Display name on the leaderboard
    value: float | str  # Numeric score or placeholder string ("***")
    order: Optional[AscendingOrder] = None  # Optional ordering preference

    @field_validator("value")
    def validate_value(cls, v):
        """
        If value is a string, it must only contain '*'.
        Example valid string: "***"
        """
        if isinstance(v, str) and any(char != "*" for char in v):
            raise ValueError(
                "If passing a string for value, it must be made entirely of the character, '*'."
            )
        return v


# -----------------------------
# JSON schema for entire assignment grading dump
# -----------------------------
class AssignmentJSON(BaseModel):
    # Top-level submission information
    score: Optional[float] = None  # Overall assignment score
    output: Optional[str] = None  # Text relevant to the whole submission
    output_format: Optional[TextFormat] = "ansi"  # Format for global output
    test_output_format: Optional[TextFormat] = (
        "ansi"  # Format for individual test outputs
    )
    visibility: Optional[Visibility] = "visible"  # Overall visibility setting
    extra_data: Optional[Dict[str, Any]] = Field(  # Arbitrary metadata
        default_factory=dict
    )

    # Per-test and leaderboard data
    tests: Optional[List[TestCaseJSON]]  # List of test case results
    leaderboard: Optional[List[LeaderboardJSON]] = None  # Leaderboard entries

    @model_validator(mode="after")
    def check_score_existence(self):
        """
        Validation logic:
        - If there are no tests, `score` must be provided.
        - If all tests are scored, `score` must NOT be overwritten.
        """
        if not self.tests:
            if self.score is None:
                raise ValueError(
                    "If there are no tests specified, please pass `score` to Grade object."
                )
        elif all(test.is_scored for test in self.tests):
            if self.score is not None:
                # Overwriting allowed by schema, but discouraged
                raise ValueError(
                    "You have specified tests with `tests`, but you are overwriting "
                    "the score of the Grade object."
                )
        return self



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, List, Mapping, Optional

from ...formatters.dispatcher import FormatDispatcher, FormatPackage
from ...json.raw_gradescope import TestCaseJSON

if TYPE_CHECKING:
    from ....types import FormatLabel
    from ...addons.addon import AddonInterface
    from ...penalties.interfaces.penalty import PenaltyInterface


class TestInterface(ABC):

    __test__: bool = False
    _created_tests: List[TestInterface] = []

    def __init__(self):
        super().__init__()
        self._penalties: List[PenaltyInterface] = []
        self._addons: List[AddonInterface] = []
        self._outputs: List[FormatPackage] = []

        self._visible: bool = True
        self._weight: float = 1.0
        self._override: Optional[bool] = None

        TestInterface._created_tests.append(self)

    @classmethod
    def clear(cls):
        cls._created_tests.clear()

    @classmethod
    def get_tests(cls) -> List[TestInterface]:
        return cls._created_tests

    @abstractmethod
    def get_name(self) -> str:
        pass

    @abstractmethod
    def get_score(self) -> float:
        pass

    def get_weight(self) -> float:
        return self._weight

    @classmethod
    def get_max_score(cls) -> float:
        return 100.0 / sum([test.get_weight() for test in cls._created_tests])

    def set_weight(self, weight: float):
        self._weight = weight

    def set_visibility(self, visibility: bool):
        self._visible = visibility

    def add_penalty(self, penalty: PenaltyInterface):
        self._penalties.append(penalty)

    def set_penalties(self, penalties: List[PenaltyInterface]):
        self._penalties = penalties

    def set_addons(self, addons: List[AddonInterface]):
        self._addons = addons

    def add_addon(self, addon: AddonInterface):
        self._addons.append(addon)

    def add_to_output(self, label: FormatLabel, data: Mapping[str, Any]):
        self._add_to_output_raw(FormatPackage(label=label, data=data))

    def get_visibility(self):
        return self._visible

    def run(self) -> TestCaseJSON:
        score: float
        if self._override is not None:
            score = self._override * self.get_weight() * self.get_max_score()
        else:
            score = self.get_score()

            for addon in self._addons:
                score *= addon.get_penalty()
                self._add_to_output_raw(addon.get_output())

            for penalty in self._penalties:
                score *= penalty.get_penalty()
                self._add_to_output_raw(penalty.get_output())

        test_output: list[str] = [
            FormatDispatcher.format(output) for output in self._outputs
        ]

        pyd_output: TestCaseJSON = TestCaseJSON(
            score=score,
            max_score=self.get_max_score(),
            name=self.get_name(),
            output="\n\n".join(test_output),
            visibility="visible" if self.get_visibility() else "hidden",
        )

        return pyd_output

    def force_pass(self) -> None:
        self._override = True

    def force_fail(self) -> None:
        self._override = False

    def _add_to_output_raw(self, package: FormatPackage):
        self._outputs.append(package)



from .dispatcher import FormatDispatcher, register_format
from .formatter import (
    ActualSTDOutFormatter,
    BuildFailureFormatter,
    ByteCmpFormatter,
    CommandFormatter,
    ExpectedSTDOutFormatter,
    MetadataFormatter,
    RawFormatter,
    STDErrFormatter,
    STDInFormatter,
    STDOutFormatter,
    StreamOutputInterface,
    UnitTestFormatter,
    ValgrindFormatter,
)

__all__ = [
    "StreamOutputInterface",
    "RawFormatter",
    "CommandFormatter",
    "STDInFormatter",
    "STDOutFormatter",
    "STDErrFormatter",
    "ActualSTDOutFormatter",
    "ExpectedSTDOutFormatter",
    "ByteCmpFormatter",
    "UnitTestFormatter",
    "BuildFailureFormatter",
    "ValgrindFormatter",
    "MetadataFormatter",
    "FormatDispatcher",
    "register_format",
]



from pathlib import Path
from typing import List, Optional

from ....data.cxx import CxxConfig
from ....data.paths import PathConfig
from ....os.file import bfs_walk, is_cxx_source_file
from ....random_utils import random_executable
from ....types import Command
from ..interfaces.cli_builder import CLIBuilderInterface


class CxxSourceBuilder(CLIBuilderInterface):

    def __init__(self):
        super().__init__()
        self._executable_path: Optional[Path] = None
        self.set_project_root(PathConfig.DEFAULT_SUBMISSION_PATH)

        self._build_directory: Optional[Path] = None
        self._built: bool = False

    def build_project(self) -> None:
        if self._built:
            return
        self._built = True
        source_files: List[Path] = []
        for file in bfs_walk(self.get_project_root()):
            if is_cxx_source_file(file):
                source_files.append(file)

        cmd: List[str | Path] = [
            "g++",
            *CxxConfig.DEFAULT_CXX_COMPILATION_FLAGS,
            f"-std={CxxConfig.DEFAULT_CXX_STANDARD}",
            "-o",
            self.get_executable_path(),
            *source_files,
        ]
        self.run_cmd(cmd)

    def get_executable_path(self) -> Path:
        if self._executable_path is not None:
            return self._executable_path
        while True:
            executable_name = self.get_build_directory() / random_executable()
            if not executable_name.exists():
                self._executable_path = executable_name
                return self._executable_path

    def set_build_directory(self, build_directory: Path) -> None:
        self._build_directory = build_directory

    def get_build_directory(self) -> Path:
        if self._build_directory is None:
            build_directory: Path = self.get_project_root() / "build"
            if not build_directory.exists():
                build_directory = self.get_project_root()
            self.set_build_directory(build_directory)
        assert self._build_directory is not None
        return self._build_directory

    def get_start_command(self) -> Command:
        if self._executable_path is None:
            self.build_project()
        assert self._executable_path is not None
        return [self._executable_path]



from datetime import datetime
from importlib import metadata
from typing import List

from pydantic import BaseModel, Field


class AssignmentMetadata(BaseModel):
    assignment_name: str
    assignment_authors: List[str]
    assignment_description: str
    assignment_due_date: datetime
    assignment_submit_date: datetime = Field(default_factory=datetime.now)

    @property
    def library_name(self) -> str:
        return "lograder"

    @property
    def library_meta(self) -> metadata.PackageMetadata:
        return metadata.metadata(self.library_name)

    @property
    def library_authors(self) -> List[str]:
        return ["Logan Dapp"]

    @property
    def library_version(self) -> str:
        return metadata.version(self.library_name)



from __future__ import annotations

import re
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, TypeGuard, cast

from ....os.cmd import run_cmd
from ....types import Command, UnitTestCase, UnitTestSuite
from ..interfaces.cli_test import CLITest
from ..interfaces.unit_test import UnitTestInterface

if TYPE_CHECKING:
    from ...unit_testers.unit_tester import UnitTesterInterface
    from ...addons.addon import ExecAddonInterface


def _is_suite(entry: UnitTestCase | UnitTestSuite) -> TypeGuard[UnitTestSuite]:
    return "cases" in entry


class Catch2UnitTest(UnitTestInterface, CLITest):
    HEADER_PATTERN = re.compile(
        r"-{79}\s*\n(.*?)\n-{79}\s*\n(.*?)(?=-{79}|={79}|$)",
        re.DOTALL,
    )

    def __init__(self):
        super().__init__()
        self._name: Optional[str] = None
        self._output: Optional[str] = None

    @classmethod
    def make(
        cls,
        name: str,
        tester: UnitTesterInterface,
        stdin: str,
        weight: float = 1.0,
        args: Optional[Command] = None,
        working_dir: Optional[Path] = None,
        wrap_args: bool = False,
    ) -> Catch2UnitTest:
        test = cls()

        test.set_name(name)
        test.set_tester(tester)
        test.set_input(stdin)
        test.set_weight(weight)

        test.set_wrap_args(wrap_args)
        if working_dir is not None:
            test.set_working_dir(working_dir)
        if args is not None:
            test.set_args(args)

        return test

    def set_name(self, name: str) -> None:
        self._name = name

    def set_output(self, output: str) -> None:
        self._output = output

    def collect_tests(self) -> UnitTestSuite:
        raw = self.get_output()

        def insert_nested_case(
            root: UnitTestSuite,
            suite_name: str,
            sections: List[str],
            success: bool,
            output: str,
        ) -> None:
            if not sections:
                case: UnitTestCase = {
                    "name": suite_name,
                    "success": success,
                    "output": output,
                }
                root["cases"].append(case)
                return

            section = sections[0]
            child_suite: UnitTestSuite | None = None

            for entry in root["cases"]:
                if _is_suite(entry) and entry["name"] == section:
                    child_suite = entry
                    break

            if child_suite is None:
                child_suite = cast(UnitTestSuite, {"name": section, "cases": []})
                root["cases"].append(child_suite)
            assert child_suite is not None
            insert_nested_case(child_suite, suite_name, sections[1:], success, output)

        top_suite: UnitTestSuite = {"name": self.get_name(), "cases": []}

        for match in self.HEADER_PATTERN.finditer(raw):
            suite_name = match.group(1).strip() or "Unnamed Suite"
            block = match.group(2).strip()

            # Look for assertion lines
            assertions = []
            for line in block.splitlines():
                if "PASSED" in line or "FAILED" in line:
                    assertions.append(line.strip())

            failed = any("FAILED" in a for a in assertions)
            sections = assertions if assertions else ["Unnamed Case"]

            insert_nested_case(
                top_suite,
                suite_name,
                sections,
                success=not failed,
                output=block,
            )

        return top_suite

    def get_args(self):
        return []

    def get_input(self):
        return ""

    def _run_test(self) -> Tuple[int, Command]:
        builder = self.get_tester()
        builder.build_project()
        if builder.get_build_error():
            self.force_fail()
            self.add_to_output("build-fail", {})
            return 1, []

        self.set_working_dir(builder.get_instance_root())

        command = builder.get_start_command() + self.get_args()

        _tmp_stdout: List[str] = []
        result = run_cmd(command, self.get_input(), [], _tmp_stdout, [])

        self._output = _tmp_stdout.pop()

        return result.returncode, command

    def get_output(self) -> str:
        if self._output is None:
            _, _ = self._run_test()
        assert self._output is not None
        return self._output

    def get_name(self) -> str:
        assert self._name is not None
        return self._name

    def add_exec_addon(self, addon: ExecAddonInterface):
        addon.set_builder(self.get_tester())
        addon.set_args(self.get_args())
        addon.set_input(self.get_input())
        self.add_addon(addon)



from .file import make_tests_from_files
from .generator import make_tests_from_generator
from .output_tests import CLIOutputTest
from .simple import make_tests_from_strs
from .template import make_tests_from_template
from .types import (
    FlaggedTestCaseProtocol,
    FlaggedWeightedTestCaseProtocol,
    TemplateSubstitution,
    TestCaseDict,
    TestCaseProtocol,
    TSub,
    WeightedTestCaseProtocol,
)

__all__ = [
    "make_tests_from_files",
    "make_tests_from_generator",
    "make_tests_from_strs",
    "make_tests_from_template",
    "CLIOutputTest",
    "TestCaseDict",
    "TemplateSubstitution",
    "TSub",
    "TestCaseProtocol",
    "FlaggedTestCaseProtocol",
    "WeightedTestCaseProtocol",
    "FlaggedWeightedTestCaseProtocol",
    "TestCaseDict",
]



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Sequence

from .output_tests import CLIOutputTest
from .validation import validate_common_size

if TYPE_CHECKING:
    from ...builders.interfaces.builder import BuilderInterface


def make_tests_from_strs(
    *,  # kwargs-only; to avoid confusion with argument sequence.
    builder: BuilderInterface,
    names: Sequence[str],
    inputs: Sequence[str],
    expected_outputs: Sequence[str],
    flag_sets: Optional[Sequence[Sequence[str]]] = None,
    weights: Optional[Sequence[float]] = None,  # Defaults to equal-weight.
) -> List[CLIOutputTest]:

    if weights is None:
        weights = [1.0 for _ in names]

    if flag_sets is None:
        flag_sets = [[] for _ in names]

    validate_common_size(
        names=names,
        inputs=inputs,
        expected_outputs=expected_outputs,
        weights=weights,
        flags=flag_sets,
    )

    generated_tests: List[CLIOutputTest] = []
    for name, input_, expected_output, weight, flags in zip(
        names, inputs, expected_outputs, weights, flag_sets, strict=True
    ):
        generated_tests.append(
            CLIOutputTest.make(
                name=name,
                builder=builder,
                stdin=input_,
                weight=weight,
                expected_stdout=expected_output,
                args=list(flags),
                working_dir=builder.get_build_directory(),
                wrap_args=builder.wrap_args(),
            )
        )
    return generated_tests



from pathlib import Path
from typing import Optional

from ....data.paths import PathConfig
from ....types import Command
from ...unit_testers.catch2 import Catch2UnitTester
from .catch2 import Catch2UnitTest


def make_unit_test_from_directory(
    *,
    name: str,
    directory: Path = Path("."),
    weight: float = 1.0,
    visible: bool = True,
    stdin: str = "",
    args: Optional[Command] = None,
    wrap_args: bool = False,
) -> Catch2UnitTest:
    tester = Catch2UnitTester()
    tester.set_testing_root(PathConfig.DEFAULT_SOURCE_PATH / directory)
    tester.set_project_root(PathConfig.DEFAULT_SUBMISSION_PATH)

    test = Catch2UnitTest.make(
        name=name,
        tester=tester,
        weight=weight,
        stdin=stdin,
        args=args,
        wrap_args=wrap_args,
    )
    test.set_visibility(visible)

    return test



from __future__ import annotations

import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

from ..data.build import BuildConfig

if TYPE_CHECKING:
    from ..types import Command


def do_process(args: Command, **kwargs) -> subprocess.CompletedProcess:
    win_prefix: Command = ["cmd", "/c"]
    cmd: Command = args
    if sys.platform.startswith("win"):
        cmd = win_prefix + cmd
    return subprocess.run(cmd, **kwargs)


def run_cmd(
    cmd: List[str | Path],
    stdin: str = "",
    commands: Optional[List[List[str | Path]]] = None,
    stdout: Optional[List[str]] = None,
    stderr: Optional[List[str]] = None,
    working_directory: Optional[Path] = None,
):
    kwargs = dict(
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        timeout=BuildConfig.DEFAULT_EXECUTABLE_TIMEOUT,
        input=stdin,
    )

    if working_directory is not None:
        kwargs["cwd"] = working_directory

    result = do_process(cmd, **kwargs)

    if commands is not None:
        commands.append(cmd)
    if stdout is not None:
        stdout.append(result.stdout)
    if stderr is not None:
        stderr.append(result.stderr)
    return result



import sys
from datetime import datetime, timedelta
from pathlib import Path

import pytest
from testing_utils import create_dummy_submission, get_results

from lograder.grader.submission_handler import SubmissionHandler
from lograder.grader.tests.unit_tests.directory import make_unit_test_from_directory


@pytest.mark.skipif(sys.platform != "win32", reason="WSL makes this really slow.")
def test_cpp_catch2_1(tmp_path: Path):
    create_dummy_submission(tmp_path, Path("tests/test-projects/cpp-catch2-project-1"))

    SubmissionHandler.clear()

    make_unit_test_from_directory(name="Catch2 Simple Project")

    SubmissionHandler.make_submission(
        assignment_name="Hello World from `lograder`!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic catch2 process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()
    for test in results["tests"]:
        assert test["score"] == test["max_score"]


@pytest.mark.skipif(sys.platform != "win32", reason="WSL makes this really slow.")
def test_cpp_catch2_2(tmp_path: Path):
    create_dummy_submission(tmp_path, Path("tests/test-projects/cpp-catch2-project-2"))

    SubmissionHandler.clear()

    make_unit_test_from_directory(name="Catch2 Simple Project")

    SubmissionHandler.make_submission(
        assignment_name="Hello World from `lograder`!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic catch2 process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()
    for test in results["tests"]:
        assert test["score"] == test["max_score"]


@pytest.mark.skipif(sys.platform != "win32", reason="WSL makes this really slow.")
def test_cpp_catch2_partial(tmp_path: Path):
    create_dummy_submission(
        tmp_path, Path("tests/test-projects/cpp-catch2-partial-project-1")
    )

    SubmissionHandler.clear()

    make_unit_test_from_directory(name="Catch2 Simple Project")

    SubmissionHandler.make_submission(
        assignment_name="Hello World from `lograder`!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic catch2 process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()
    for test in results["tests"]:
        assert test["score"] == test["max_score"] * 0.5


@pytest.mark.skipif(sys.platform != "win32", reason="WSL makes this really slow.")
def test_cmake_catch2_1(tmp_path: Path):
    create_dummy_submission(
        tmp_path, Path("tests/test-projects/cmake-catch2-project-1")
    )

    SubmissionHandler.clear()

    make_unit_test_from_directory(name="Catch2 Simple Project")

    SubmissionHandler.make_submission(
        assignment_name="Hello World from `lograder`!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic catch2 process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()
    for test in results["tests"]:
        assert test["score"] == test["max_score"]


@pytest.mark.skipif(sys.platform != "win32", reason="WSL makes this really slow.")
def test_cmake_catch2_2(tmp_path: Path):
    create_dummy_submission(
        tmp_path, Path("tests/test-projects/cmake-catch2-project-2")
    )

    SubmissionHandler.clear()

    make_unit_test_from_directory(name="Catch2 Simple Project")

    SubmissionHandler.make_submission(
        assignment_name="Hello World from `lograder`!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic catch2 process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()
    for test in results["tests"]:
        assert test["score"] == test["max_score"]



from __future__ import annotations

from abc import ABC
from pathlib import Path
from typing import List, Optional

from ....os.cmd import run_cmd
from ....types import Command, CommandOutput, StreamOutput
from ...submission_handler import SubmissionHandler
from ..interfaces.builder import BuilderInterface


class CLIBuilderInterface(BuilderInterface, ABC):
    def __init__(self):
        super().__init__()

        self._commands: List[List[str | Path]] = []
        self._stdout: List[str] = []
        self._stderr: List[str] = []

    def get_commands(self) -> List[List[str | Path]]:
        return self._commands

    def get_stdout(self) -> List[str]:
        return self._stdout

    def get_stderr(self) -> List[str]:
        return self._stderr

    def run_cmd(self, cmd: Command, working_directory: Optional[Path] = None) -> None:
        result = run_cmd(
            cmd,
            commands=self._commands,
            stdout=self._stdout,
            stderr=self._stderr,
            working_directory=working_directory,
        )
        SubmissionHandler.add_output(
            "command",
            CommandOutput(
                command=self._commands[-1],
                exit_code=result.returncode,
            ),
        )
        SubmissionHandler.add_output(
            "stdout", StreamOutput(stream_contents=self._stdout[-1])
        )
        SubmissionHandler.add_output(
            "stderr", StreamOutput(stream_contents=self._stderr[-1])
        )
        if result.returncode != 0:
            self.set_build_error(True)



from __future__ import annotations

import shlex
from abc import ABC
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from .test import TestInterface

if TYPE_CHECKING:
    from ....types import Command


class CLITest(TestInterface, ABC):
    def __init__(self):
        super().__init__()

        self._args: Command = []
        self._wrap_args: bool = False
        self._working_dir: Optional[Path] = None
        self._stdin: str = ""

    def set_input(self, stdin: str) -> None:
        self._stdin = stdin

    def get_input(self) -> str:
        return self._stdin

    def set_working_dir(self, path: Path):
        self._working_dir = path

    def get_working_dir(self) -> Optional[Path]:
        return self._working_dir

    def set_wrap_args(self, wrap: bool = True) -> None:
        self._wrap_args = wrap

    def get_wrap_args(self) -> bool:
        return self._wrap_args

    def set_args(self, args: Command) -> None:
        self._args = args

    def get_args(self) -> Command:
        if self.get_wrap_args():
            return [
                f'ARGS="{shlex.join([str(arg.resolve()) if isinstance(arg, Path) else arg for arg in self._args])}"'
            ]
        return self._args



from typing import Any, Sequence


def validate_common_size(**seqs: Sequence[Any]):
    seq_lens = [len(seq) for _, seq in seqs.items()]
    if not seq_lens:
        return
    initial_seq_len = seq_lens[0]
    if not all([initial_seq_len == seq_len for seq_len in seq_lens]):
        raise MismatchedSequenceLengthError(**seqs)


def validate_unique_argument(**kwargs: Any):
    if sum([val is None for val in kwargs.values()]) != 1:
        raise NonSingleArgumentSpecifiedError(**kwargs)


class LograderValidationError(Exception):
    """
    This is the base exception class for all exceptions raised
    by the `lograder.tests` module, whenever a validation error
    occurs for easy error handling.
    """


class MismatchedSequenceLengthError(LograderValidationError):
    """
    This is the exception that is raised when the inputted
    sequence lengths to a function do not match.
    """

    def __init__(self, **seqs: Sequence[Any]):
        super().__init__(
            "Mismatched sequence lengths passed to parameters: "
            + ", ".join([f"`{kw}` (length of {len(seq)})" for kw, seq in seqs.items()])
        )


class NonSingleArgumentSpecifiedError(LograderValidationError):
    """
    This is the exception that is raised when more than one or zero
    arguments are passed.
    """

    def __init__(self, **kwargs: Any):
        super().__init__(
            "Must specify only a single argument; please choose one. Received the arguments: "
            + ", ".join(
                [f"`{kw}` = {val}" for kw, val in kwargs.items() if val is not None]
            )
        )


class ArgumentSpecifiedError(LograderValidationError):
    """
    This is the exception that is raised when an argument is passed
    when it should not have been.
    """

    def __init__(self, conflicting_arg: str, **specified_args: Any):
        super().__init__(
            "Specified arguments, "
            + ", ".join(
                [
                    f"`{kw}` = {val}"
                    for kw, val in specified_args.items()
                    if val is not None
                ]
            )
            + f", that should have been left blank because argument, `{conflicting_arg}`, was specified."
        )



import json
import shutil
from pathlib import Path

from lograder.data.paths import PathConfig


def get_results() -> dict:
    return json.load(open(PathConfig.DEFAULT_RESULT_PATH))


def create_dummy_submission(grader_root: Path, project_root: Path) -> None:
    root_path = grader_root / "autograder"
    submission_path = root_path / "submission"
    source_path = root_path / "source"
    result_path = root_path / "results"
    json_path = result_path / "results.json"

    root_path.mkdir(parents=True, exist_ok=True)
    submission_path.mkdir(parents=True, exist_ok=True)
    source_path.mkdir(parents=True, exist_ok=True)
    result_path.mkdir(parents=True, exist_ok=True)

    PathConfig.DEFAULT_ROOT_PATH = root_path
    PathConfig.DEFAULT_SUBMISSION_PATH = submission_path
    PathConfig.DEFAULT_SOURCE_PATH = source_path
    PathConfig.DEFAULT_RESULT_PATH = json_path

    sub_root = project_root / "submission"
    sou_root = project_root / "source"
    if sub_root.exists() and sou_root.exists():
        shutil.copytree(sub_root, submission_path, dirs_exist_ok=True)
        shutil.copytree(sou_root, source_path, dirs_exist_ok=True)
    else:
        shutil.copytree(project_root, submission_path, dirs_exist_ok=True)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, List, Literal, TypedDict, Union, cast

if TYPE_CHECKING:
    from .grader.addons.valgrind import ValgrindLeakSummary, ValgrindWarningSummary
    from .grader.metadata import AssignmentMetadata

# --- BUILD TYPES ---
Command = List[Union[str, Path]]
ProjectType = Literal["cmake", "cxx-source", "makefile", "py-source", "pyproject"]
TestType = Literal["catch2", "pytest"]

# --- GRADESCOPE TYPES ---
TextFormat = Literal["text", "html", "simple_format", "md", "ansi"]
Visibility = Literal["visible", "after_due_date", "after_published", "hidden"]
Status = Literal["passed", "failed"]
AscendingOrder = Literal["asc"]

# --- FORMATTER TYPES ---
FormatLabel = Literal[
    "assignment-metadata",
    "command",
    "raw",
    "valgrind",
    "stdin",
    "expected-stdout",
    "actual-stdout",
    "byte-cmp",
    "stderr",
    "stdout",
    "unit-tests",
    "build-fail",
]


class StreamOutput(TypedDict):
    stream_contents: str


class ByteStreamComparisonOutput(TypedDict):
    stream_actual_bytes: bytes
    stream_expected_bytes: bytes


class ValgrindOutput(TypedDict):
    leaks: ValgrindLeakSummary
    warnings: ValgrindWarningSummary


class CommandOutput(TypedDict):
    command: Command
    exit_code: int


class AssignmentMetadataOutput(TypedDict):
    metadata: AssignmentMetadata


# --- UNIT-TEST TYPES ---
class UnitTestCase(TypedDict):
    name: str
    success: bool
    output: str


class UnitTestSuite(TypedDict):
    name: str
    cases: List[UnitTestCase | UnitTestSuite]


def is_successful_test(suite: UnitTestSuite | UnitTestCase) -> bool:
    if "success" in suite.keys():
        return cast(UnitTestCase, suite)["success"]
    return all(is_successful_test(case) for case in cast(UnitTestSuite, suite)["cases"])



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Sequence

from .simple import make_tests_from_strs
from .types import TemplateSubstitution
from .validation import ArgumentSpecifiedError, validate_unique_argument

if TYPE_CHECKING:
    from ...builders.interfaces.builder import BuilderInterface
    from .output_tests import CLIOutputTest


class TestCaseTemplate:
    __test__: bool = False

    def __init__(
        self,
        *,
        input_strs: Optional[Sequence[str]] = None,
        input_template_file: Optional[Path] = None,
        input_template_str: Optional[str] = None,
        input_substitutions: Optional[Sequence[TemplateSubstitution]] = None,
        expected_output_strs: Optional[Sequence[str]] = None,
        expected_output_template_file: Optional[Path] = None,
        expected_output_template_str: Optional[str] = None,
        expected_output_substitutions: Optional[Sequence[TemplateSubstitution]] = None,
        flag_sets: Optional[Sequence[List[str]]] = None,
    ):

        if input_strs is not None:
            if (
                input_template_file is not None
                or input_template_str is not None
                or input_substitutions is not None
            ):
                raise ArgumentSpecifiedError(
                    "input_strs",
                    input_template_file=input_template_file,
                    input_template_str=input_template_str,
                    input_substitutions=input_substitutions,
                )
        else:
            validate_unique_argument(
                input_template_file=input_template_file,
                input_template_str=input_template_str,
            )
            if input_template_str is None:
                assert input_template_file is not None
                with open(Path(input_template_file), "r") as f:
                    input_template_str = f.read()

            assert input_substitutions is not None
            input_strs = [
                input_template_str.format(*temp.args, **temp.kwargs)
                for temp in input_substitutions
            ]

        if expected_output_strs is not None:
            if (
                expected_output_template_file is not None
                or expected_output_template_str is not None
                or expected_output_substitutions is not None
            ):
                raise ArgumentSpecifiedError(
                    "expected_output_strs",
                    expected_output_template_file=expected_output_template_file,
                    expected_output_template_str=expected_output_template_str,
                    expected_output_substitutions=expected_output_substitutions,
                )
        else:
            validate_unique_argument(
                expected_output_template_file=expected_output_template_file,
                expected_output_template_str=expected_output_template_str,
            )
            if expected_output_template_str is None:
                assert expected_output_template_file is not None
                with open(Path(expected_output_template_file), "r") as f:
                    expected_output_template_str = f.read()

            assert expected_output_substitutions is not None
            expected_output_strs = [
                expected_output_template_str.format(*temp.args, **temp.kwargs)
                for temp in expected_output_substitutions
            ]

        self._inputs = input_strs
        self._expected_outputs = expected_output_strs
        self._flag_sets = flag_sets

    def get_flag_sets(self):
        return self._flag_sets

    def get_inputs(self):
        return self._inputs

    def get_outputs(self):
        return self._expected_outputs


def make_tests_from_template(
    builder: BuilderInterface,
    names: Sequence[str],
    template: TestCaseTemplate,
    weights: Optional[Sequence[float]] = None,
) -> List[CLIOutputTest]:
    return make_tests_from_strs(
        builder=builder,
        names=names,
        inputs=template.get_inputs(),
        expected_outputs=template.get_outputs(),
        weights=weights,
        flag_sets=template.get_flag_sets(),
    )



from __future__ import annotations

from typing import List, TYPE_CHECKING
from datetime import datetime

from lograder.grader.submission_handler import SubmissionHandler
from lograder.grader.tests.unit_tests.directory import make_unit_test_from_directory
from lograder.grader.builders.dispatcher import ProjectDispatcher
from lograder.grader.builders.cpp.cxx_source import CxxSourceBuilder
from lograder.grader.builders.cpp.cmake import CMakeBuilder
from lograder.grader.tests.output_comparison import (
    TestCaseDict,
    TestCaseProtocol,
    make_tests_from_files,
    make_tests_from_generator,
    make_tests_from_strs,
    make_tests_from_template,
    TemplateSubstitution,
    TSub,
)
from lograder.grader.addons.valgrind import ValgrindAddon
from lograder.grader.tests.interfaces.test import TestInterface

if TYPE_CHECKING:
    from lograder.grader.tests.output_comparison.output_tests import OutputTestInterface
    from lograder.grader.tests.unit_tests.catch2 import Catch2UnitTest

__all__ = [
    "make_unit_test_from_directory",
    "make_tests_from_files",
    "make_tests_from_generator",
    "make_tests_from_strs",
    "make_tests_from_template",
    "ProjectDispatcher",
    "CMakeBuilder",
    "CxxSourceBuilder",
    "TestCaseDict",
    "TestCaseProtocol"
]

def initialize_assignment():
    SubmissionHandler.clear()

def add_valgrind(test: OutputTestInterface | Catch2UnitTest):
    test.add_exec_addon(ValgrindAddon())

def get_all_tests() -> List[TestInterface]:
    return TestInterface.get_tests()

def finalize_assignment(
    assignment_name: str,
    assignment_authors: List[str],
    assignment_description: str,
    assignment_due_date: datetime
):
    SubmissionHandler.make_submission(
        assignment_name=assignment_name,
        assignment_authors=assignment_authors,
        assignment_description=assignment_description,
        assignment_due_date=assignment_due_date,
    )

def hello_world():
    return "Hello World from `lograder`!"



from pathlib import Path
from typing import List, Optional

from ...data.paths import PathConfig
from ...os.file import detect_project_type
from ...types import Command, ProjectType
from .cpp.cmake import CMakeBuilder
from .cpp.cxx_source import CxxSourceBuilder
from .interfaces.cli_builder import CLIBuilderInterface


class ProjectDispatcher(CLIBuilderInterface):
    def __init__(self):
        super().__init__()
        self._builder: Optional[CLIBuilderInterface] = None
        self._allowed_project_types: List[ProjectType] = [
            "cmake",
            "cxx-source",
            "makefile",
            "py-source",
            "pyproject",
        ]
        self.set_project_root(PathConfig.DEFAULT_SUBMISSION_PATH)

    def set_allowed_project_types(self, allowed_project_types: List[ProjectType]):
        self._allowed_project_types = allowed_project_types

    def load_builder(self):
        if self._builder is not None:
            return
        project_type = detect_project_type(self.get_project_root())
        if project_type not in self._allowed_project_types:
            self.set_build_error(True)
            return

        if project_type == "cmake":
            self._builder = CMakeBuilder()
        else:
            self._builder = CxxSourceBuilder()

    def get_builder(self) -> CLIBuilderInterface:
        if self._builder is None:
            self.load_builder()
        assert self._builder is not None
        return self._builder

    def get_build_directory(self) -> Path:
        return self.get_builder().get_build_directory()

    def get_start_command(self) -> Command:
        return self.get_builder().get_start_command()

    def build_project(self) -> None:
        self.get_builder().build_project()



from pathlib import Path
from typing import Optional

from ...os.file import bfs_walk, is_cmake_file
from ...types import Command
from ..builders.cpp.cmake import CMakeBuilder
from ..builders.cpp.cxx_source import CxxSourceBuilder
from ..builders.interfaces.builder import BuilderInterface
from .unit_tester import UnitTesterInterface


class Catch2UnitTester(UnitTesterInterface):
    def __init__(self):
        super().__init__()
        self._builder: Optional[BuilderInterface] = None

    def get_builder(self) -> BuilderInterface:
        assert self._builder is not None
        return self._builder

    def get_build_directory(self) -> Path:
        return self.get_instance_root()

    def get_start_command(self) -> Command:
        return self.get_builder().get_start_command() + ["--success"]

    def build_test(self) -> None:
        cmake: bool = False
        for file in bfs_walk(self.get_instance_root()):
            if is_cmake_file(file):
                cmake = True
                break
        if cmake:
            self._builder = CMakeBuilder()
        else:
            self._builder = CxxSourceBuilder()

        self._builder.set_project_root(self.get_instance_root())
        self._builder.build_project()



import re
import sys
from pathlib import Path
from typing import List, Optional

from ....data.cxx import CxxConfig
from ....data.paths import PathConfig
from ....os.file import bfs_walk, is_cmake_file, is_valid_target
from ....random_utils import random_name
from ....types import Command
from ..interfaces.cli_builder import CLIBuilderInterface


class CMakeBuilder(CLIBuilderInterface):

    TARGET_PATTERN = re.compile(r"^\.\.\.\s+([a-zA-Z0-9_\-.]+)", re.MULTILINE)
    WIN_TARGET_PATTERN = re.compile(r"^([A-Za-z0-9_\-.]+):", re.MULTILINE)

    def __init__(self):
        super().__init__()
        self._executable_path: Optional[Path] = None
        self._project_root: Path = Path(PathConfig.DEFAULT_SUBMISSION_PATH)

        self._build_directory: Path = self._project_root / f"build-{random_name()}"
        self._build_directory.mkdir(parents=True, exist_ok=True)
        self._working_directory: Optional[Path] = None
        self._cmake_file: Optional[Path] = None
        self._target: Optional[str] = None

    def set_working_directory(self, path: Path) -> None:
        self._working_directory = path

    def get_working_directory(self) -> Path:
        assert self._working_directory is not None
        return self._working_directory

    def find_executable(self, target: str) -> Path:
        build_dir = self.get_build_directory()
        candidates = [
            build_dir / "Debug" / f"{target}.exe",
            build_dir / "Release" / f"{target}.exe",
            build_dir / "Debug" / target,
            build_dir / "Release" / target,
            build_dir / f"{target}.exe",
            build_dir / target,
        ]

        for path in candidates:
            if path.is_file():
                return path

        for path in build_dir.rglob("*"):
            if path.is_file():
                if path.name == target or path.name == f"{target}.exe":
                    return path

        raise FileNotFoundError

    def build_project(self) -> None:
        for file in bfs_walk(self._project_root):
            if is_cmake_file(file):
                self._cmake_file = file
                break
        assert self._cmake_file is not None
        self.set_working_directory(self._cmake_file.parent)

        if sys.platform.startswith("win"):
            cmd: List[str | Path] = [
                "cmake",
                "-S",
                self.get_working_directory(),
                "-B",
                self.get_build_directory(),
                "-G",
                "Ninja",
            ]
        else:
            cmd: List[str | Path] = [
                "cmake",
                "-S",
                self.get_working_directory(),
                "-B",
                self.get_build_directory(),
                "-G",
                "Unix Makefiles",
            ]

        self.run_cmd(cmd)
        if self.get_build_error():
            return

        cmd = ["cmake", "--build", self.get_build_directory(), "--target", "help"]
        self.run_cmd(cmd)
        if self.get_build_error():
            return

        if sys.platform.startswith("win"):
            targets = self.WIN_TARGET_PATTERN.findall(self.get_stdout()[-1])
        else:
            targets = self.TARGET_PATTERN.findall(self.get_stdout()[-1])

        if "main" in targets:
            self._target = "main"
        elif "build" in targets:
            self._target = "build"
        elif "demo" in targets:
            self._target = "demo"
        else:
            valid_targets = [t for t in targets if is_valid_target(t)]
            if not valid_targets:
                self.set_build_error(True)
                return
            self._target = valid_targets[0]

        print(self._target)

        if sys.platform.startswith("win"):
            cmd = [
                "cmake",
                *CxxConfig.DEFAULT_CMAKE_COMPILATION_FLAGS,
                "--build",
                self.get_build_directory(),
                "--target",
                self._target,
                "--",
                "--quiet",
            ]
        else:
            cmd = [
                "cmake",
                *CxxConfig.DEFAULT_CMAKE_COMPILATION_FLAGS,
                "--build",
                self.get_build_directory(),
                "--target",
                self._target,
                "--",
                "-s",
                "--no-print-directory",
            ]
        self.run_cmd(cmd)
        if self.get_build_error():
            return

        self._executable_path = self.find_executable(self._target)
        try:
            self._executable_path = self.find_executable(self._target)
        except FileNotFoundError:
            self.set_build_error(True)
            return

        assert self._executable_path is not None

    def set_build_directory(self, build_directory: Path) -> None:
        self._build_directory = build_directory

    def get_build_directory(self) -> Path:
        return self._build_directory

    def get_start_command(self) -> Command:
        if self._executable_path is None:
            self.build_project()
        if self.get_build_error():
            return []
        assert self._executable_path is not None
        return [self._executable_path]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from ....data.messages import MessageConfig
from .test import TestInterface

if TYPE_CHECKING:
    from ....types import ByteStreamComparisonOutput, StreamOutput


class OutputTestInterface(TestInterface, ABC):

    def __init__(self):
        super().__init__()
        self._run: bool = False
        self._stdin: str = ""

    def set_input(self, stdin: str) -> None:
        self._stdin = stdin

    def get_input(self) -> str:
        return self._stdin

    @abstractmethod
    def get_expected_output(self) -> str:
        pass

    @abstractmethod
    def get_actual_output(self) -> str:
        pass

    @abstractmethod
    def get_error(self) -> str:
        pass

    def get_score(self) -> float:
        input: StreamOutput = {"stream_contents": self.get_input()}
        expected: StreamOutput = {"stream_contents": self.get_expected_output().strip()}
        actual: StreamOutput = {"stream_contents": self.get_actual_output().strip()}
        error: StreamOutput = {"stream_contents": self.get_error()}

        byte_cmp: ByteStreamComparisonOutput = {
            "stream_expected_bytes": expected["stream_contents"].encode(
                "ascii", errors="ignore"
            ),
            "stream_actual_bytes": actual["stream_contents"].encode(
                "ascii", errors="ignore"
            ),
        }

        if not self._run:  # stop duplicate appending
            self.add_to_output("stdin", input)
            self.add_to_output("stderr", error)
            self.add_to_output("expected-stdout", expected)
            self.add_to_output("actual-stdout", actual)
            self.add_to_output(
                "raw", {"stream_contents": MessageConfig.DEFAULT_RAW_TITLE}
            )
            self.add_to_output("byte-cmp", byte_cmp)
            self._run = True

        return (
            (expected["stream_contents"] == actual["stream_contents"])
            * self.get_max_score()
            * self.get_weight()
        )



import sys
from datetime import datetime, timedelta
from pathlib import Path

import pytest
from testing_utils import create_dummy_submission, get_results

from lograder.data.penalties import PenaltyConfig
from lograder.grader.addons.valgrind import ValgrindAddon
from lograder.grader.builders.dispatcher import ProjectDispatcher
from lograder.grader.submission_handler import SubmissionHandler
from lograder.grader.tests.output_comparison import make_tests_from_strs


@pytest.mark.skipif(sys.platform == "win32", reason="Does not run on Windows")
def test_valgrind_no_leak(tmp_path: Path):
    create_dummy_submission(
        tmp_path, Path("tests/test-projects/cpp-valgrind-project-1")
    )

    SubmissionHandler.clear()

    builder = ProjectDispatcher()
    builder.set_allowed_project_types(["cxx-source"])

    tests = make_tests_from_strs(
        builder=builder,
        names=["Test `Hello World`."],
        inputs=[""],
        expected_outputs=[""],
    )
    tests[0].add_exec_addon(ValgrindAddon())
    SubmissionHandler.make_submission(
        assignment_name="Hello World from `lograder`!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic compilation process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()
    for test in results["tests"]:
        assert test["score"] == test["max_score"]


@pytest.mark.skipif(sys.platform == "win32", reason="Does not run on Windows")
def test_valgrind_with_leak(tmp_path: Path):
    create_dummy_submission(
        tmp_path, Path("tests/test-projects/cpp-valgrind-project-2")
    )

    SubmissionHandler.clear()

    builder = ProjectDispatcher()
    builder.set_allowed_project_types(["cxx-source"])

    tests = make_tests_from_strs(
        builder=builder,
        names=["Test `Hello World`."],
        inputs=[""],
        expected_outputs=[""],
    )
    tests[0].add_exec_addon(ValgrindAddon())
    SubmissionHandler.make_submission(
        assignment_name="Hello World from `lograder`!",
        assignment_authors=["Logan Dapp"],
        assignment_description="Test the most basic compilation process.",
        assignment_due_date=datetime.now() + timedelta(hours=9.53),
    )

    results = get_results()
    for test in results["tests"]:
        assert (
            test["score"]
            == test["max_score"] * PenaltyConfig.DEFAULT_VALGRIND_LEAK_PENALTY
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

from ..penalties.interfaces.penalty import PenaltyInterface

if TYPE_CHECKING:
    from ...types import Command
    from ..builders.interfaces.builder import BuilderInterface


class AddonInterface(PenaltyInterface, ABC):
    def __init__(self):
        super().__init__()
        self._builder: Optional[BuilderInterface] = None

    def get_builder(self) -> BuilderInterface:
        assert self._builder is not None
        return self._builder

    def set_builder(self, builder: BuilderInterface) -> None:
        self._builder = builder

    @abstractmethod
    def run(self) -> None:
        pass


class ExecAddonInterface(AddonInterface, ABC):
    def __init__(self):
        super().__init__()
        self._stdin: str = ""
        self._args: Command = []

    def set_args(self, args: Command):
        self._args = args

    def get_args(self) -> Command:
        return self._args

    def set_input(self, stdin: str):
        self._stdin = stdin

    def get_input(self):
        return self._stdin



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List, Optional, cast

from ....types import UnitTestCase, UnitTestSuite
from .test import TestInterface

if TYPE_CHECKING:
    from ...unit_testers.unit_tester import UnitTesterInterface


class UnitTestInterface(TestInterface, ABC):

    def __init__(self):
        super().__init__()
        self._tester: Optional[UnitTesterInterface] = None
        self._run: bool = False
        self._stdin: str = ""

    def set_input(self, stdin: str) -> None:
        self._stdin = stdin

    def get_input(self) -> str:
        return self._stdin

    @abstractmethod
    def collect_tests(self) -> UnitTestSuite:
        pass

    def get_score(self) -> float:
        tests = self.collect_tests()
        if not self._run:
            self.add_to_output("unit-tests", tests)
            self._run = True

        cases: List[UnitTestCase] = []
        suites: List[UnitTestSuite] = [tests]
        while suites:
            suite = suites.pop()
            for test in suite["cases"]:
                if "success" in test.keys():
                    cases.append(cast(UnitTestCase, test))
                else:
                    suites.append(cast(UnitTestSuite, test))

        outputs: List[bool] = [test["success"] for test in cases]
        tests_passed: int = sum(outputs)
        tests_total: int = max(len(outputs), 1)

        return (tests_passed / tests_total) * self.get_max_score() * self.get_weight()

    def get_tester(self) -> UnitTesterInterface:
        assert self._tester is not None
        return self._tester

    def set_tester(self, tester: UnitTesterInterface):
        self._tester = tester



from datetime import datetime
from typing import Any, List, Mapping

from ..data.paths import PathConfig
from ..types import AssignmentMetadataOutput, FormatLabel
from .formatters.dispatcher import FormatDispatcher, FormatPackage
from .json.raw_gradescope import AssignmentJSON, TestCaseJSON
from .metadata import AssignmentMetadata
from .tests.interfaces.test import TestInterface


class SubmissionHandler:

    _outputs: List[FormatPackage] = []

    @classmethod
    def add_output(cls, label: FormatLabel, data: Mapping[str, Any]):
        cls._outputs.append(FormatPackage(label=label, data=data))

    @classmethod
    def make_submission(
        cls,
        assignment_name: str,
        assignment_authors: List[str],
        assignment_description: str,
        assignment_due_date: datetime,
    ) -> AssignmentJSON:
        metadata: AssignmentMetadata = AssignmentMetadata(
            assignment_name=assignment_name,
            assignment_authors=assignment_authors,
            assignment_description=assignment_description,
            assignment_due_date=assignment_due_date,
        )
        packed_metadata: AssignmentMetadataOutput = {"metadata": metadata}
        if not cls._outputs or cls._outputs[0].label != "assignment-metadata":
            cls._outputs.insert(
                0, FormatPackage(label="assignment-metadata", data=packed_metadata)
            )
        build_output: list[str] = [
            FormatDispatcher.format(output) for output in cls._outputs
        ]

        assignment = AssignmentJSON(
            output="\n\n".join(build_output), tests=cls.run_tests()
        )

        with open(PathConfig.DEFAULT_RESULT_PATH, "w", encoding="utf-8") as f:
            f.write(assignment.model_dump_json(indent=2))

        return assignment

    @classmethod
    def clear(cls):
        cls._outputs.clear()
        TestInterface.clear()

    @classmethod
    def run_tests(cls) -> List[TestCaseJSON]:
        return [t.run() for t in TestInterface.get_tests()]
